#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from enum import Enum

from synnax.exceptions import ValidationError


class FramingMode(Enum):
    KEY = "key"
    NAME = "name"
    UNOPENED = "unopened"


def open_framing_mode(
    keys: list[str] | None = None,
    names: list[str] | None = None,
) -> FramingMode:
    if keys is not None and names is not None:
        raise ValidationError(
            "keys and names cannot both be specified when calling open()"
        )
    elif keys is None and names is None:
        raise ValidationError("keys or names must be specified when calling open()")
    return FramingMode.KEY if keys is not None else FramingMode.NAME



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from freighter import (
    URL,
    AsyncMiddleware,
    AsyncStreamClient,
    HTTPClientFactory,
    JSONEncoder,
    Middleware,
    MsgpackEncoder,
    StreamClient,
    SyncStreamClient,
    WebsocketClient,
)


class Transport:
    url: URL
    stream: StreamClient
    stream_async: AsyncStreamClient
    http: HTTPClientFactory
    secure: bool

    def __init__(self, url: URL, secure: bool = False) -> None:
        self.url = url.child("/api/v1/")
        self.stream_async = WebsocketClient(
            base_url=self.url,
            encoder=MsgpackEncoder(),
            max_message_size=int(5e6),
            secure=secure,
        )
        self.stream = SyncStreamClient(self.stream_async)
        self.http = HTTPClientFactory(
            url=self.url,
            encoder_decoder=JSONEncoder(),
            secure=secure,
        )

    def use(self, *middleware: Middleware):
        self.http.use(*middleware)
        self.stream.use(*middleware)

    def use_async(self, *middleware: AsyncMiddleware):
        self.stream_async.use(*middleware)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from numpy import ndarray
import pandas as pd

from synnax.channel.retrieve import ChannelRetriever
from synnax.telem import NumpyArray, TimeRange, UnparsedTimeStamp
from synnax.transport import Transport
from synnax.framer.iterator import AUTO_SPAN, NumpyIterator
from synnax.framer.writer import DataFrameWriter


class FrameClient:
    """SegmentClient provides interfaces for reading and writing segmented
    telemetry from a Synnax Cluster. SegmentClient should not be instantiated
    directly, but rather used through the synnax.Synnax class.
    """

    _transport: Transport
    _channels: ChannelRetriever

    def __init__(self, transport: Transport, registry: ChannelRetriever):
        self._transport = transport
        self._channels = registry

    def new_writer(
        self,
        start: UnparsedTimeStamp,
        keys: list[str] | None = None,
        names: list[str] | None = None,
        strict: bool = False,
        suppress_warnings: bool = False,
    ) -> DataFrameWriter:
        """Opens a new writer on the given channels.

        :param keys: A list of channel keys that the writer will write to. A writer
        cannot write to keys not provided in this list. See the NumpyWriter documentation
        for more.
        :returns: A NumpyWriter that can be used to write telemetry to the given channels.
        """
        w = DataFrameWriter(
            client=self._transport.stream,
            channels=self._channels,
            strict=strict,
            suppress_warnings=suppress_warnings,
        )
        w.open(start, keys, names)
        return w

    def new_iterator(
        self,
        tr: TimeRange,
        keys: list[str] | None = None,
        names: list[str] | None = None,
        aggregate: bool = False,
    ) -> NumpyIterator:
        """Opens a new iterator over the given channels within the provided time range.

        :param keys: A list of channel keys to iterator over.
        :param tr: A time range to iterate over.
        :param aggregate:  Whether to accumulate iteration results or reset them on every
        iterator method call.
        :returns: A NumpyIterator over the given channels within the provided time
        range. See the NumpyIterator documentation for more.
        """
        npi = NumpyIterator(
            transport=self._transport.stream,
            channels=self._channels,
            aggregate=aggregate,
        )
        npi.open(tr, keys, names)
        return npi

    def write(
        self,
        start: UnparsedTimeStamp,
        data: ndarray,
        key: str | None = None,
        name: str | None = None,
        strict: bool = False,
    ):
        """Writes telemetry to the given channel starting at the given timestamp.

        :param to: The key of the channel to write to.
        :param start: The starting timestamp of the first sample in data.
        :param data: The telemetry to write to the channel.
        :returns: None.
        """
        to = key if key else name
        w = self.new_writer(
            start=start,
            keys=[key] if key else None,
            names=[name] if name else None,
            strict=strict,
        )
        try:
            w.write(pd.DataFrame({to: data}))
            w.commit()
        finally:
            w.close()

    def read(
        self,
        start: UnparsedTimeStamp,
        end: UnparsedTimeStamp,
        key: str | None = None,
        name: str | None = None,
    ) -> tuple[ndarray, TimeRange]:
        """Reads telemetry from the channel between the two timestamps.

        :param from_: THe key of the channel to read from.
        :param start: The starting timestamp of the range to read from.
        :param end: The ending timestamp of the range to read from.
        :returns: A numpy array containing the retrieved telemetry.
        :raises ContiguityError: If the telemetry between start and end is non-contiguous.
        """
        arr = self.read_array(start, end, key, name)
        assert arr.time_range is not None
        return arr.data, arr.time_range

    def read_array(
        self,
        start: UnparsedTimeStamp,
        end: UnparsedTimeStamp,
        key: str | None = None,
        name: str | None = None,
    ) -> NumpyArray:
        """Reads a Segment from the given channel between the two timestamps.

        :param from_: The key of the channel to read from.
        :param start: The starting timestamp of the range to read from.
        :param end: The ending timestamp of the range to read from.
        :returns: A NumpySegment containing the read telemetry.
        :raises ContiguityError: If the telemetry between start and end is non-contiguous.
        """
        from_ = key if key else name
        i = self.new_iterator(
            tr=TimeRange(start, end),
            aggregate=True,
            keys=[key] if key else None,
            names=[name] if name else None,
        )
        try:
            i.seek_first()
            while i.next(AUTO_SPAN):
                pass
        finally:
            i.close()
        return i.value[from_]



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from enum import Enum

from freighter import EOF, ExceptionPayload, Payload, Stream, StreamClient

from pandas import DataFrame

from synnax.exceptions import UnexpectedError, GeneralError
from synnax.framer.mode import FramingMode, open_framing_mode
from synnax.telem import TimeRange, TimeSpan, TimeStamp
from synnax.framer.payload import BinaryFrame, NumpyFrame
from synnax.channel.retrieve import ChannelRetriever

AUTO_SPAN = TimeSpan(-1)


class _Command(int, Enum):
    OPEN = 0
    NEXT = 1
    PREV = 2
    SEEK_FIRST = 3
    SEEK_LAST = 4
    SEEK_LE = 5
    SEEK_GE = 6
    VALID = 7
    ERROR = 8


class _ResponseVariant(int, Enum):
    NONE = 0
    ACK = 1
    DATA = 2


class _Request(Payload):
    command: _Command
    span: TimeSpan | None = None
    range: TimeRange | None = None
    stamp: TimeStamp | None = None
    keys: list[str] | None = None


class _Response(Payload):
    variant: _ResponseVariant
    command: _Command
    ack: bool
    error: ExceptionPayload
    frame: BinaryFrame


class CoreIterator:
    """Used to iterate over a databases telemetry in time-order. It should not be
    instantiated directly, and should instead be instantiated using the segment Client.

    Using an iterator is ideal when querying/processing large ranges of data, but is
    relatively complex and difficult to use. If you're looking to retrieve telemetry
    between two timestamps, see the segment Client read method instead.
    """

    _ENDPOINT = "/frame/iterate"

    aggregate: bool
    _client: StreamClient
    _stream: Stream[_Request, _Response]
    _value: BinaryFrame

    def __init__(self, client: StreamClient, aggregate: bool = False) -> None:
        self._client = client
        self.aggregate = aggregate

    def open(self, tr: TimeRange, keys: list[str]):
        """Opens the iterator, configuring it to iterate over the telemetry in the
        channels with the given keys within the provided time range.

        :param keys: The keys of the channels to iterate over.
        :param tr: The time range to iterate over.
        """
        self._stream = self._client.stream(self._ENDPOINT, _Request, _Response)
        self._exec(command=_Command.OPEN, range=tr, keys=keys)
        self._value = BinaryFrame()

    def next(self, span: TimeSpan) -> bool:
        """Reads the next time span of telemetry for each channel in the iterator.

        :param span: The span of time to read. A negative span is equivalent to calling
        prev with the absolute value of the span. If AUTO_SPAN is provided, the iterator
        will automatically determine the span of time to read. This is useful for iterating
        quickly over an entire time range.

        :returns: False if a segment satisfying the request can't be found for a
        particular channel or the iterator has accumulated an error.
        """
        return self._exec(command=_Command.NEXT, span=span)

    def prev(self, span: TimeSpan) -> bool:
        """Reads the previous time span of telemetry for each channel in the iterator.

        :param span: The span of time to read. A negative span is equivalent to calling
        next with the absolute value of the span. If AUTO_SPAN is provided, the iterator
        will automatically determine the span of time to read. This is useful for iterating
        quickly over an entire time range.

        :returns: False if a segment satisfying the request can't be found for a particular
        channel or the iterator has accumulated an error.
        """
        return self._exec(command=_Command.NEXT, span=span)

    def seek_first(self) -> bool:
        """Seeks the iterator to the first segment in the time range, but does not read
        it. Also invalidates the iterator. The iterator will not be considered valid
        until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_FIRST)

    def seek_last(self) -> bool:
        """Seeks the iterator to the last segment in the time range, but does not read it.
        Also invalidates the iterator. The iterator will not be considered valid
        until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_LAST)

    def seek_lt(self, stamp: TimeStamp) -> bool:
        """Seeks the iterator to the first segment whose start is less than or equal to
        the provided timestamp. Also invalidates the iterator. The iterator will not be
        considered valid until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_LE, stamp=stamp)

    def seek_ge(self, stamp: TimeStamp) -> bool:
        """Seeks the iterator to the first segment whose start is greater than or equal to
        the provided timestamp. Also invalidates the iterator. The iterator will not be
        considered valid until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_GE, stamp=stamp)

    def valid(self) -> bool:
        """Returns true if the iterator value contains a valid segment, and False otherwise.
        valid most commonly returns false when the iterator is exhausted or has accumulated
        an error.
        """
        return self._exec(command=_Command.VALID)

    def close(self):
        """Close closes the iterator. An iterator MUST be closed after use, and this method
        should probably be placed in a 'finally' block. If the iterator is not closed, it make
        leak resources and threads.
        """
        exc = self._stream.close_send()
        if exc is not None:
            raise exc
        _, exc = self._stream.receive()
        if exc is None:
            raise UnexpectedError(
                """Unexpected missing close acknowledgement from server.
                Please report this issue to the Synnax team."""
            )
        elif not isinstance(exc, EOF):
            raise exc

    @property
    def value(self) -> BinaryFrame:
        return self._value.compact()

    def _exec(self, **kwargs) -> bool:
        exc = self._stream.send(_Request(**kwargs))
        if exc is not None:
            raise exc
        if not self.aggregate:
            self._value = BinaryFrame()
        while True:
            r, exc = self._stream.receive()
            if exc is not None:
                raise exc
            assert r is not None
            if r.variant == _ResponseVariant.ACK:
                return r.ack
            self._value.append_frame(r.frame)


class NumpyIterator(CoreIterator):
    """Used to iterate over a databases telemetry in time-order. It should not be
    instantiated directly, and should instead be instantiated using the segment Client.

    Using an iterator is ideal when querying/processing large ranges of data, but is
    relatively complex and difficult to use. If you're looking to retrieve telemetry
    between two timestamps, see the segment Client read method instead.
    """

    _channels: ChannelRetriever
    _mode: FramingMode

    def __init__(
        self,
        transport: StreamClient,
        channels: ChannelRetriever,
        aggregate: bool = False,
    ):
        super().__init__(transport, aggregate)
        self._channels = channels

    def open(
        self,
        tr: TimeRange,
        keys: list[str] | None = None,
        names: list[str] | None = None,
    ):
        self._mode = open_framing_mode(keys, names)
        channels = self._channels.retrieve(keys=keys, names=names, node_id=None)
        super().open(tr, [ch.key for ch in channels])

    @property
    def value(self) -> NumpyFrame:
        """
        :returns: The current iterator value as a dictionary whose keys are channels
        and values are segments containing telemetry at the current iterator position.
        """
        v = super().value
        v.keys = self._value_keys(v)
        return NumpyFrame.from_binary(v)

    def _value_keys(self, fr: BinaryFrame) -> list[str]:
        keys = [k for v in fr for k in v]
        if self._mode == FramingMode.KEY:
            return keys
        if self._mode == FramingMode.NAME:
            # We can safely ignore the none case here because we've already
            # checked that all channels can be retrieved.
            return [self._channels.retrieve(k).name for k in keys]  # type: ignore
        raise GeneralError(
            "Must call open() before attempting to access iterator value."
        )



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import overload, Literal
from numpy import ndarray
from pydantic import PrivateAttr

from synnax.exceptions import ValidationError, QueryError
from synnax.framer import FrameClient
from synnax.telem import (
    Rate,
    Density,
    TimeRange,
    UnparsedDataType,
    UnparsedDensity,
    UnparsedRate,
    UnparsedTimeStamp,
    DataType,
)

from synnax.channel.create import ChannelCreator
from synnax.channel.payload import ChannelPayload
from synnax.channel.retrieve import ChannelRetriever


class Channel(ChannelPayload):
    """Represents a Channel in a Synnax database."""

    __frame_client: FrameClient | None = PrivateAttr(None)

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self,
        data_type: UnparsedDataType,
        rate: UnparsedRate = 0,
        name: str = "",
        node_id: int = 0,
        key: str = "",
        is_index: bool = False,
        index: str = "",
        density: UnparsedDensity = 0,
        frame_client: FrameClient | None = None,
    ):
        super().__init__(
            data_type=DataType(data_type),
            rate=Rate(rate),
            name=name,
            node_id=node_id,
            key=key,
            density=Density(density),
            is_index=is_index,
            index=index,
        )
        self.__frame_client = frame_client

    def _payload(self) -> ChannelPayload:
        return ChannelPayload(
            data_type=self.data_type,
            density=self.density,
            rate=self.rate,
            name=self.name,
            node_id=self.node_id,
            key=self.key,
            index=self.index,
            is_index=self.is_index,
        )

    def read(
        self, start: UnparsedTimeStamp, end: UnparsedTimeStamp
    ) -> tuple[ndarray, TimeRange]:
        """Reads telemetry from the channel between the two timestamps.

        :param start: The starting timestamp of the range to read from.
        :param end: The ending timestamp of the range to read from.
        :returns: A numpy array containing the retrieved telemetry from the database.
        :raises ContiguityError: If the telemetry between start and end is non-contiguous.
        """
        return self._frame_client.read(start, end, key=self.key)

    def write(self, start: UnparsedTimeStamp, data: ndarray):
        """Writes telemetry to the channel starting at the given timestamp.

        :param start: The starting timestamp of the first sample in data.
        :param data: The telemetry to write to the channel.
        :returns: None.
        """
        self._frame_client.write(start, data, key=self.key)

    @property
    def _frame_client(self) -> FrameClient:
        if self.__frame_client is None:
            raise ValidationError(
                "Cannot read from a channel that has not been created."
            )
        return self.__frame_client

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return self.key == other.key

    def __str__(self):
        base = f"{self.name} ({self.data_type})"
        if self.rate != 0:
            base += f" @ {self.rate}Hz"
        return base


class ChannelClient:
    """The core client class for executing channel operations against a Synnax cluster."""

    _frame_client: FrameClient
    _retriever: ChannelRetriever
    _creator: ChannelCreator

    def __init__(
        self,
        frame_client: FrameClient,
        retriever: ChannelRetriever,
        creator: ChannelCreator,
    ):
        self._frame_client = frame_client
        self._retriever = retriever
        self._creator = creator

    @overload
    def create(
        self,
        channels: None = None,
        data_type: UnparsedDataType = DataType.UNKNOWN,
        name: str = "",
        node_id: int = 0,
        rate: UnparsedRate = Rate(0),
        index: str = "",
        is_index: bool = False,
    ) -> Channel:
        """Creates a channel using the given template.

        :param name: The name of the channel to create.
        :param rate: The sample rate of the channel in Hz.
        :param data_type: The data type of the channel. Can be any type in
        UnparsedDataType, such as np.float64 or np.int64,
        :param node_id: The node that holds the lease on the channel. If you don't know
        what this is, don't worry about it.
        :returns: The created channel.
        """
        ...

    @overload
    def create(self, channels: Channel) -> Channel:
        ...

    @overload
    def create(self, channels: list[Channel]) -> list[Channel]:
        ...

    def create(
        self,
        channels: Channel | list[Channel] | None = None,
        data_type: UnparsedDataType = DataType.UNKNOWN,
        name: str = "",
        node_id: int = 0,
        rate: UnparsedRate = Rate(0),
        index: str = "",
        is_index: bool = False,
    ) -> Channel | list[Channel]:
        if channels is None:
            _channels = [
                ChannelPayload(
                    name=name,
                    node_id=node_id,
                    rate=Rate(rate),
                    data_type=DataType(data_type),
                    index=index,
                    is_index=is_index,
                )
            ]
        elif isinstance(channels, Channel):
            _channels = [channels._payload()]
        else:
            _channels = [c._payload() for c in channels]
        created = self._sugar(self._creator.create(_channels))
        return created if isinstance(channels, list) else created[0]

    @overload
    def retrieve(self, keys: str | None = None, names: str | None = None) -> Channel:
        ...

    @overload
    def retrieve(
        self,
        keys: list[str] | None = None,
        names: list[str] | None = None,
        node_id: int | None = None,
        include_not_found: Literal[False] = False,
    ) -> list[Channel]:
        ...

    @overload
    def retrieve(
        self,
        keys: list[str] | None = None,
        names: list[str] | None = None,
        node_id: int | None = None,
        include_not_found: Literal[True] = True,
    ) -> tuple[list[Channel], list[str]]:
        ...

    def retrieve(
        self,
        keys: str | list[str] | None = None,
        names: str | list[str] | None = None,
        node_id: int | None = None,
        include_not_found: bool = False,
    ) -> Channel | list[Channel] | tuple[list[Channel], list[str]]:
        res = self._retriever.retrieve(
            keys=keys, names=names, node_id=node_id, include_not_found=include_not_found
        )
        if res is None:
            raise QueryError("No channels found.")
        if isinstance(res, ChannelPayload):
            return self._sugar([res])[0]
        if isinstance(res, tuple):
            return self._sugar(res[0]), res[1]
        return self._sugar(res)

    def _sugar(self, channels: list[ChannelPayload]) -> list[Channel]:
        return [Channel(**c.dict(), frame_client=self._frame_client) for c in channels]



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Any

from rich import print
from rich.prompt import Confirm, FloatPrompt, IntPrompt, Prompt

from synnax.cli.console.protocol import Console

SYNNAX_PRIMARY_Z = "#3774D0"


class RichConsole:
    """A Console implementation using Rich."""

    info_color: str
    warn_color: str
    error_color: str
    success_color: str

    def __init__(
        self,
        info_color: str = SYNNAX_PRIMARY_Z,
        warn_color: str = "yellow",
        error_color: str = "red",
        success_color: str = "green",
    ):
        self.info_color = info_color
        self.warn_color = warn_color
        self.error_color = error_color
        self.success_color = success_color

    def _(self) -> Console:
        return self

    def info(self, message: str) -> None:
        print(f"[{self.info_color}]{message}[/]")

    def warn(self, message: str) -> None:
        print(f"[{self.warn_color}]{message}[/]")

    def error(self, message: str) -> None:
        print(f"[{self.error_color}]{message}[/]")

    def success(self, message: str) -> None:
        print(f"[{self.success_color}]{message}[/]")

    def ask(
        self,
        question: str,
        choices: list[str] | None = None,
        default: str | None = None,
        required: bool = False,
    ) -> str | None:
        res = Prompt.ask(
            question,
            choices=choices,
            default=default,
        )
        if self._check_required(required, res):
            return self.ask(question, choices, default, required)
        return res

    def ask_int(
        self,
        question: str,
        bound: tuple[int, int] | None = None,
        default: int | None = None,
        required: bool = False,
    ) -> int | None:
        res = IntPrompt.ask(
            question,
            default=default,
            choices=[str(i) for i in range(*bound)] if bound else None,
        )
        if self._check_required(required, res):
            return self.ask_int(question, bound, default, required)
        return res

    def ask_float(
        self,
        question: str,
        default: float | None = None,
        required: bool = False,
    ) -> float | None:
        res = FloatPrompt.ask(
            question,
            default=default,
        )
        if self._check_required(required, res):
            return self.ask_float(question, default, required)
        return res

    def ask_password(
        self,
        question: str,
        required: bool = False,
    ) -> str:
        res = Prompt.ask(
            question,
            password=True,
        )
        if self._check_required(required, res):
            return self.ask_password(question, required)
        return res

    def confirm(
        self,
        question: str,
        default: bool = True,
    ) -> bool:
        return Confirm.ask(
            question,
            default=default,
        )

    def table(
        self,
        columns: list[str],
        rows: list[dict],
    ) -> None:
        from rich.table import Table

        table = Table(show_header=True, header_style="bold magenta")
        for column in columns:
            table.add_column(column)
        for row in rows:
            # order the row.py by the columns
            table.add_row(*[row[column] for column in columns])

        print(table)

    def _warn_required(self) -> None:
        self.warn("This is a required field.")

    def _check_required(self, required: bool, res: Any | None) -> bool:
        if (res is None or res == "") and required:
            self._warn_required()
            return True
        return False



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations
from typing import Protocol, overload
from typing_extensions import Literal

from freighter import HTTPClientFactory, Payload, UnaryClient

from synnax.exceptions import QueryError
from synnax.channel.payload import ChannelPayload


class _Request(Payload):
    keys: list[str] | None = None
    node_id: int | None = None
    names: list[str] | None = None


class _Response(Payload):
    channels: list[ChannelPayload] = []
    not_found: list[str] = []


class ChannelRetriever(Protocol):
    @overload
    def retrieve(
        self,
        keys: str | None = None,
        names: str | None = None,
    ) -> ChannelPayload | None:
        ...

    @overload
    def retrieve(
        self,
        keys: list[str] | None = None,
        names: list[str] | None = None,
        node_id: int | None = None,
        include_not_found: Literal[False] = False,
    ) -> list[ChannelPayload]:
        ...

    @overload
    def retrieve(
        self,
        keys: list[str] | None = None,
        names: list[str] | None = None,
        node_id: int | None = None,
        include_not_found: Literal[True] = True,
    ) -> tuple[list[ChannelPayload], list[str]]:
        ...

    @overload
    def retrieve(
        self,
        keys: str | list[str] | None = None,
        names: str | list[str] | None = None,
        node_id: int | None = None,
        include_not_found: bool = False,
    ) -> list[ChannelPayload] | tuple[
        list[ChannelPayload], list[str]
    ] | ChannelPayload | None:
        ...


class ClusterChannelRetriever:
    _ENDPOINT = "/channel/retrieve"
    client: UnaryClient

    def __init__(self, client: HTTPClientFactory):
        self.client = client.get_client()

    def _(self) -> ChannelRetriever:
        return self

    def retrieve(
        self,
        keys: str | list[str] | None = None,
        names: str | list[str] | None = None,
        node_id: int | None = None,
        include_not_found: bool = False,
    ) -> tuple[list[ChannelPayload], list[str]] | list[
        ChannelPayload
    ] | ChannelPayload | None:
        single_key = isinstance(keys, str)
        single_name = isinstance(names, str)
        req = _Request(
            keys=[keys] if single_key else keys,
            names=[names] if single_name else names,
            node_id=node_id,
        )
        res, exc = self.client.send(self._ENDPOINT, req, _Response)
        if exc is not None:
            raise exc
        assert res is not None
        if include_not_found is True:
            return res.channels, res.not_found
        if single_key or single_name:
            if len(res.channels) == 1:
                return res.channels[0]
            if len(res.channels) == 0:
                return None
            raise QueryError("multiple channels found")
        return res.channels


class CacheChannelRetriever:
    retriever: ChannelRetriever
    channels: dict[str, ChannelPayload]
    names_to_keys: dict[str, str]

    def __init__(self, retriever: ChannelRetriever) -> None:
        self.channels = dict()
        self.names_to_keys = dict()
        self.retriever = retriever

    def _(self) -> ChannelRetriever:
        return self

    def retrieve(
        self,
        keys: str | list[str] | None = None,
        names: str | list[str] | None = None,
        node_id: int | None = None,
        include_not_found: bool = False,
    ) -> tuple[list[ChannelPayload], list[str]] | list[
        ChannelPayload
    ] | ChannelPayload | None:
        if node_id is not None:
            return self.retriever.retrieve(
                node_id=node_id, 
                include_not_found=include_not_found
            )

        keys, single_key = self._normalize(keys)
        names, single_name = self._normalize(names)
        keys_to_retrieve = list()
        names_to_retrieve = list()
        results = list()

        for name in names:
            key = self.names_to_keys.get(name, None)
            if key is not None:
                keys.append(key)
            else:
                names_to_retrieve.append(name)

        for key in keys:
            channel = self.channels.get(key, None)
            if channel is None:
                keys_to_retrieve.append(key)
            else:
                results.append(channel)

        if len(keys_to_retrieve) == 0 and len(names_to_retrieve) == 0:
            return results

        retrieved, not_found = self.retriever.retrieve(
            keys=keys_to_retrieve,
            names=names_to_retrieve,
            include_not_found=True,
        )

        for channel in retrieved:
            self.channels[channel.key] = channel
            self.names_to_keys[channel.name] = channel.key
            results.append(channel)

        if include_not_found:
            return results, not_found
        if single_key or single_name:
            return None if len(results) == 0 else results[0]
        return results

    def _normalize(self, keys: list[str] | str | None) -> tuple[list[str], bool]:
        if keys is None:
            return [], False
        if isinstance(keys, str):
            return [keys], True
        return keys, False



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import TextIO
from pydantic import BaseModel

from synnax.cli.console.protocol import Print, Prompt

Response = str | int | float | bool | None


class Entry(BaseModel):
    message: str | None = None
    columns: list[str] | None = None
    rows: list[dict] | None = None
    choices: list[str] | None = None
    default: Response = None
    response: Response = None
    required: bool | None = None


class Output(BaseModel):
    entries: list[Entry]

    def __init__(self, entries: list[Entry] | None):
        self.entries = entries or list()

    def append(self, entry: Entry):
        assert self.entries is not None
        self.entries.append(entry)

    def write(self, f: TextIO):
        f.write(self.json())


class MockPrint(Print):
    """A mock implementation of the Print protocol for testing purposes."""

    output: Output

    def __init__(self, output: Output):
        """
        :param output: The output list to append entries to.
        """
        self.output = output

    def _(self) -> Print:
        return self

    def info(self, message: str):
        self.output.append(Entry(message=message))

    def error(self, message: str):
        self.output.append(Entry(message=message))

    def warn(self, message: str):
        self.output.append(Entry(message=message))

    def success(self, message: str):
        self.output.append(Entry(message=message))

    def table(self, columns: list, rows: list):
        self.output.append(Entry(columns=columns, rows=rows))


class MockPrompt(Prompt):
    """A mock implementation of the Prompt protocol for testing purposes."""

    output: Output
    responses: list

    def __init__(self, output: Output, responses: list):
        """
        :param output: The output list to append entries to.
        :param responses: A list of responses to return in order. These responses
        must be valid for the type of prompt being used.
        """
        self.output = output
        self.responses = responses

    def _(self) -> Prompt:
        return self

    def ask(
        self,
        question: str,
        choices: list[str] | None = None,
        default: str | None = None,
        required: bool = False,
    ):
        e = Entry(message=question, choices=choices, default=default, required=required)
        e.response = self.responses.pop(0) or default
        if type(e.response) != str:
            raise TypeError("Unexpected response type for ask. It should be a string.")
        return e.response

    def ask_int(
        self, question: str, default: int | None = None, required: bool = False
    ):
        e = Entry(message=question, default=default, required=required)
        e.response = self.responses.pop(0) or default
        if type(e.response) != int:
            raise TypeError(
                "Unexpected response type for ask_int. It should be an integer."
            )
        return default

    def ask_float(
        self, question: str, default: float | None = None, required: bool = False
    ):
        e = Entry(message=question, default=default, required=required)
        e.response = self.responses.pop(0) or default
        if type(e.response) != float:
            raise TypeError(
                "Unexpected response type for ask_float. It should be a float."
            )
        return e.response

    def ask_password(self, message: str, required: bool):
        e = Entry(message=message, required=required)
        e.response = self.responses.pop(0)
        if e.response is None:
            raise ValueError("Password cannot be empty.")
        elif type(e.response) != str:
            raise TypeError(
                "Unexpected response type for ask_password. It should be a string."
            )
        return e.response

    def confirm(self, message: str, default: bool = True):
        e = Entry(message=message, default=default)
        e.response = self.responses.pop(0) or default
        if type(e.response) != bool:
            raise TypeError(
                "Unexpected response type for confirm. It should be a boolean."
            )
        return e.response


class MockConsole(MockPrint, MockPrompt):
    """A mock implementation of the Console protocol for testing purposes."""

    def __init__(self, output: Output, responses: list):
        """
        :param output: The output list to append entries to.
        :param responses: A list of responses to return in order. These responses
        must be valid for the type of prompt being used.
        """
        self.output = output
        self.responses = responses



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.cli.flow import Context
from typing import Any


def select_from_table(
    ctx: Context,
    columns: list[str],
    rows: list[dict[str, Any]],
    default: int | None = None,
    required: bool = True,
) -> int | None:
    """Prompts the user to select a row from a table.

    :param ctx: The current flow Context.
    :param columns: The columns of the table.
    :param rows: The rows of the table.
    :param allow_none: Whether to allow the user to select nothing.
    :param default: The default option to select. If a default is provided,
    allow_none is ignored.
    :returns: The index of the selected row.py or None if nothing was selected.
    """
    ctx.console.table(
        columns=["option", *columns],
        rows=[{"option": str(i), **row} for i, row in enumerate(rows)],
    )
    if required and default is None:
        ctx.console.info("Press enter to select nothing.")
    i = ctx.console.ask_int("Select an option #", bound=(0, len(rows)), default=default)
    return None if i == "None" else i


def select_simple(
    ctx: Context,
    choices: list[str],
    default: int | None = None,
    required: bool = True,
) -> int | None:
    """Prompts the user to select a choice from the given choices.

    :param ctx: The current flow Context.
    :param choices: The choices to select from.
    :param default: The default option to select. If a default is provided,
    allow_none is ignored.
    :returns: The index of the selected choice or None if nothing was selected.
    """
    return select_from_table(
        ctx,
        ["choice"],
        [{"choice": choice} for choice in choices],
        default,
        required,
    )



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from synnax.telem import DataType, TimeSpan
from synnax.cli.channel import select_from_table
from synnax.cli.flow import Context


def prompt_data_type_select(ctx: Context, required: bool = True) -> DataType | None:
    """Prompts the user to select a data type from a list of all available data
    types.

    :param ctx: The current flow Context.
    :param allow_none: Whether to allow the user to select None.
    """
    i = select_from_table(
        ctx,
        ["data_type"],
        [{"data_type": name.string()} for name in DataType.ALL],
        required,
    )
    return DataType.ALL[i] if i is not None else None


def ask_time_units_select(
    ctx: Context, required: bool = True, question: str | None = None
) -> str | None:
    """Prompts the user to select a time unit from a list of all available time
    units.

    :param ctx: The current flow Context.
    :param allow_none: Whether to allow the user to select None.
    """
    if question is not None:
        ctx.console.ask(question)
    opts = list(TimeSpan.UNITS.keys())
    i = select_from_table(
        ctx,
        ["unit"],
        [{"unit": unit} for unit in opts],
        required,
    )
    return opts[i] if i is not None else None



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from pathlib import Path

import click
import numpy as np

from synnax.io import BaseReader
from synnax.io.factory import IO_FACTORY
from synnax.telem import convert_time_units
from synnax.cli.flow import Context
from synnax.cli.io import prompt_new_reader
from synnax.cli.telem import ask_time_units_select
from synnax.cli.args import if_none
from synnax.cli import default


@click.command()
@click.argument("path", required=False)
@click.option("-ip", "--input", required=False)
@click.option("-op", "--output", required=False)
@click.option("-o", "--out", required=False)
@click.option("-c", "--channel", required=False)
def tsconvert(
    path: str | None,
    out: str | None,
    channel: str | None,
    input: str | None,
    output: str | None,
) -> None:
    pure_tsconvert(path, out, channel, input, output)


def pure_tsconvert(
    path: Path | str | None,
    out: Path | str | None,
    channel: str | None,
    input: str | None,
    output: str | None,
    ctx: Context = default.context(),
) -> None:
    reader = prompt_new_reader(ctx, path)
    if reader is None:
        return

    channel = ask_channel_and_check_exists(
        ctx,
        reader,
        channel,
        "Which channel would you like to convert?",
    )
    if channel is None:
        return

    input = if_none(
        input,
        ask_time_units_select,
        ctx,
        question="What is the current precision?",
        required=True,
    )
    output = if_none(
        output,
        ask_time_units_select,
        ctx,
        question="What is the desired precision?",
        required=True,
    )
    out = if_none(
        out,
        ctx.console.ask,
        "Where would you like to save the converted data?",
        required=True,
    )
    assert output is not None and input is not None and out is not None
    writer = IO_FACTORY.new_writer(reader.path().parent / out)
    reader.seek_first()
    for chunk in reader:
        chunk[channel] = convert_time_units(chunk[channel], input, output).astype(
            np.int64
        )
        writer.write(chunk)


def ask_channel_and_check_exists(
    ctx: Context,
    reader: BaseReader,
    channel: str | None,
    question="Enter a channel name",
) -> str | None:
    assert reader is not None
    _ch = if_none(
        channel,
        ctx.console.ask,
        question,
        required=True,
    )
    try:
        next(ch for ch in reader.channels() if ch.name == _ch)
    except StopIteration:
        ctx.console.error(f"Channel not found: {_ch}")
        if channel is None:
            return ask_channel_and_check_exists(ctx, reader, channel, question)
    return _ch



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import TypeVar, Any, Callable

V = TypeVar("V")


def if_none(
    value: V | None,
    func: Callable[[Any], V],
    *args: Any,
    **kwargs: Any,
) -> V:
    """If value is None, return the result of calling func with args."""
    return func(*args, **kwargs) if value is None else value
